/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#ifndef DATAFOX_COMMUNICATION_CONNECTION_BASE_H
#define DATAFOX_COMMUNICATION_CONNECTION_BASE_H

// Standardtypen
#include <cstdint>

/////////////////////////////
// Error's
#include "common/error_defines.h"

#ifdef _MSC_VER
#define __func__ __FUNCTION__
#pragma warning(disable : 4996)
#endif

#if defined(_MSC_VER) || defined(__MINGW32__)


typedef int socklen_t;

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <winsock2.h>
#include <string>

#ifndef STRING2
#define STRING2(x) #x
#define STRING(x) STRING2(x)
#define TODO(x) __pragma(message(__FILE__ "(" STRING(__LINE__) ") : todo : " #x))
#define FIXME(x) __pragma(message(__FILE__ "(" STRING(__LINE__) ") : fixme : " #x))
#endif

#else

#include <string>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
typedef int HANDLE;
#define SOCKET int
#define MAX_PATH 260
#define _MAX_PATH MAX_PATH
#define _MAX_EXT 256
#define TRUE 1
#define FALSE 0
#include <termio.h>
#include <sys/file.h>

#define INVALID_HANDLE_VALUE -1

#include <errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define INVALID_SOCKET       -1
#define SOCKADDR_IN struct sockaddr_in
#define SOCKADDR struct sockaddr
#define HOSTENT struct hostent
#define WSAGetLastError() h_errno
#define SOCKET_ERROR -1

#define FAR

#ifndef STRING2
#define STRING2(x) #x
#define STRING(x) STRING2(x)
#define TODO(x)
#define FIXME(x)
#endif

#endif

#include <memory>

namespace Datafox
{
    namespace Communication
    {
        class Trace_base
        {
        public:
            enum class Trace_mode
            {
                off,
                on,
                packets
            };

        public:
            Trace_base( int id );

            void print( const char *format, va_list args );

            inline int get_id()
            {
                return m_id;
            }
            inline void set_id( int id )
            {
                m_id = id;
            }

            inline Trace_mode get_mode()
            { 
                return m_mode; 
            }
            inline void set_mode( Trace_mode mode = Trace_mode::on )
            { 
                m_mode = mode;
            }

            inline std::string get_file_path()
            {
                return m_file_path;
            }
            inline void set_file_path( const std::string &file_path )
            { 
                m_file_path = file_path;
            }

        private:
            int m_id;
            Trace_mode m_mode;            // Mode des Trace
            std::string m_file_path;      // Dateiangabe zum Trace
        };

        // Basisklasse fuer pure Kommunikationsaufgaben.
        class Connection_base  
        {
        public:
            enum // default value
            {
                dv_connection_id = 0,
                dv_read_trials   = 3, // Maximale Anzahl Leseversuche
                dv_write_trials  = 3  // Maximale Anzahl Schreibversuche
            };

        public:
            Connection_base( int id = dv_connection_id );
            virtual ~Connection_base();

        public:
            // Typ der Verbindung
            enum class Connection_type 
            { 
                undefined, 
                serial, 
                socket
            };

            // Oeffnen der Schnittstelle
            virtual int open( Connection_type type, const std::string &connection_string, uint32_t connection_value, uint32_t communication_read_timeout );
            // Schliessen der Schnittstelle.
            virtual int close();

            // Zuweisen von HANDLE oder Socket ohne Schnittstelle selbst zu oeffnen.
            int attach( Connection_type type, HANDLE connection_handle, uint32_t communication_read_timeout );
            // Freigeben der anderweitig zugewiesenen Resourcen
            int detach();

            // Ermitteln ob Schnittstelle geoeffnet.
            bool is_open();

            // Ermitteln der uebergebenen Parameter an Open.
            inline void get_open_parameters( Connection_type &type, std::string &connection_string, uint32_t &connection_value, uint32_t &communication_read_timeout ) 
            {
                type = m_type;
                connection_string = m_string;
                connection_value = m_value;
                communication_read_timeout = m_read_timeout;
            }

        public:
            // Loeschen anstehender Daten.
            void flush();
            int flush_socket();

        public:
            // Oeffnen der Schnittstelle fuer RS232 / RS485 Verbindung.
            int open_serial( const std::string &serial_port_name, uint32_t baud_rate, uint32_t communication_read_timeout );
            int close_serial();
            int change_state_serial( HANDLE connection_serial_port, uint32_t baud_rate, uint32_t communication_read_timeout );

            // Oeffnen der Schnittstelle fuer TCP/IP.
            int open_socket( const std::string &host_name, uint16_t port_number, uint32_t communication_read_timeout );
            int close_socket();

        protected:
            // Datenempfang
            int read_serial( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials = dv_read_trials );
            int read_socket( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials = dv_read_trials );

        public:
            virtual int read_data( uint8_t *buffer, std::size_t read_count, std::size_t &read_in_count, uint8_t read_trials = dv_read_trials );

        protected:
            // Datenversand
            int write_serial( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials = dv_write_trials );
            int write_socket( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials = dv_write_trials );

        public:
            virtual int write_data( const uint8_t *buffer, std::size_t write_count, std::size_t &write_out_count, uint8_t write_trials = dv_write_trials );

        private:
            std::shared_ptr<Trace_base> m_trace;

        public:
            std::shared_ptr<Trace_base> get_trace_pointer()
            {
                return m_trace;
            }
            void set_trace_pointer( std::shared_ptr<Trace_base> base )
            {
                m_trace = base;
            }

        public:
            virtual void trace( const char *format... )
            {
                if ( m_trace.get() != nullptr && m_trace->get_mode() >= Trace_base::Trace_mode::on )
                {
                    va_list args;
                    va_start( args, format );
                    m_trace->print( format, args );
                    va_end( args );
                }
            }

            Trace_base::Trace_mode get_trace_mode()
            {
                Trace_base::Trace_mode mode = Trace_base::Trace_mode::off;
                if ( m_trace.get() != nullptr )
                {
                    mode = m_trace->get_mode();
                }

                return mode;
            }
            void set_trace_mode( Trace_base::Trace_mode mode = Trace_base::Trace_mode::on )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->set_mode( mode );
                }
            }

            std::string get_trace_file_path()
            {
                std::string file_path;
                if ( m_trace.get() != nullptr )
                {
                    file_path = m_trace->get_file_path();
                }

                return file_path;
            }
            void set_trace_file_path( const std::string &file_path )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->set_file_path( file_path );
                }
            }

        public:
            inline int get_id() 
            { 
                return m_id; 
            }
            inline int set_id( int id ) 
            { 
                int old = m_id;
                m_id = id; 
                if ( m_trace.get() != nullptr )
                {
                    m_trace->set_id( id );
                }
                return old; 
            }
        public:
            inline Connection_type get_type() 
            { 
                return m_type; 
            }
            inline void set_type( Connection_type type )
            {
                m_type = type;
            }
        public:
            inline HANDLE get_serial_port()
            {
                return m_serial_port;
            }
            inline void set_serial_port( HANDLE serial_port )
            {
                m_serial_port = serial_port;
            }
        public:
            inline SOCKET get_socket()
            {
                return m_socket;
            }
            inline void set_socket( SOCKET socket )
            {
                m_socket = socket;
            }
        public:
            inline std::string get_host_name()
            {
                return m_string;
            }
            inline void set_host_name( const std::string &name )
            {
                m_string = name;
            }
        public:
            inline std::string get_serial_port_name()
            {
                return m_string;
            }
            inline void set_serial_port_name( const std::string &name )
            {
                m_string = name;
            }
        public:
            inline uint16_t get_port_number()
            {
                return (uint16_t)m_value;
            }
            inline void set_port_number( uint16_t value )
            {
                m_value = value;
            }
        public:
            inline uint32_t get_baud()
            {
                return m_value;
            }
            inline void set_baud( uint32_t value )
            {
                m_value = value;
            }
        public:
            inline uint32_t get_read_timeout() 
            { 
                return m_read_timeout; 
            }
            inline void set_read_timeout( uint32_t timeout )
            { 
                m_read_timeout = timeout; 
            }

        protected:
            int m_id;                      // Id des Channels
            Connection_type m_type;        // Typ der Verbindung / Schnittstelle
            HANDLE m_serial_port;          // Handle wenn Typ: ct_seriel
            SOCKET m_socket;               // Socket wenn Typ: ct_socket
            std::string m_string;          // Haelt letzten ComPort-String
            uint32_t m_value;              // Baudrate bei Typ Seriell, Portnummer bei Typ Socket
            uint32_t m_read_timeout;       // Lesetimeout fuer Kommunikationsfunktionen
#if !defined(_MSC_VER) && !defined(__MINGW32__)
            struct termios m_termio;
            struct flock  m_flock;
#endif

        protected:
            std::size_t m_size_written;    // Statistik, Anzahl der versendeten Bytes
            std::size_t m_size_read;       // Statistik, Anzahl der gelesenen Bytes

        public:
            inline std::size_t get_size_written()
            {
                return m_size_written;
            }
            inline void set_size_written( std::size_t size )
            {
                m_size_written = size;
            }
            inline void add_size_written( std::size_t size )
            {
                m_size_written += size;
            }
            inline void sub_size_written( std::size_t size )
            {
                m_size_written -= size;
            }

            inline std::size_t get_size_read()
            {
                return m_size_read;
            }
            inline void set_size_read( std::size_t size )
            {
                m_size_read = size;
            }
            inline void add_size_read( std::size_t size )
            {
                m_size_read += size;
            }
            inline void sub_size_read( std::size_t size )
            {
                m_size_read -= size;
            }
        };

    } // namespace Communication
} // namespace Datafox

#endif
